/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include <stdlib.h>

#ifndef NeXT
# include <unistd.h>
#endif

# include <errno.h>
# include <fcntl.h>
# ifdef sequent
#   include <sys/types.h>
# endif
# include <sys/stat.h>

# ifndef TRUE
#   define TRUE 1
#   define FALSE 0
# endif

static char *filebuf = 0;
static char *lastptr = 0;
static char *ptr = 0;
static int nolines = 0;
static char **lines = 0;

void openFile (char *path)
{
  int i;
  int infd;
  char *ptr;
  struct stat statbuf;

  if ( filebuf ) { free (filebuf); filebuf = 0; }
  if ( lines   ) { free (lines  ); lines   = 0; }

  if ( (infd = open (path, O_RDONLY)) == -1 ) {
    perror (path);
    exit (1);
  }
  if ( fstat (infd, &statbuf) == -1 ) {
    fprintf (stderr, "%s: cannot determine size\n", path);
    exit (1);
  }

  if ( (filebuf = malloc (statbuf.st_size)) == NULL ) {
    fprintf (stderr, "%s: cannot get enough memory\n", path);
    exit (1);
  }

  if ( read (infd, (void *) filebuf, statbuf.st_size) != statbuf.st_size ) {
    fprintf (stderr, "%s: cannot read\n", path);
    exit (1);
  }
  close (infd);

  nolines = 0;
  for (i=0; i<statbuf.st_size; i++) {
    if ( filebuf[i] == '\n' ) {
      filebuf[i] = '\0';
      nolines++;
    }
  }

  if ( (lines = malloc (nolines * sizeof(char *))) == NULL ) {
    fprintf (stderr, "%s: cannot setup line buffer\n", path);
    exit (1);
  }

  nolines = 0;
  ptr     = filebuf;
  for (i=0; i<statbuf.st_size; i++) {
    if ( filebuf[i] == '\0' ) {
      lines[nolines++] = ptr;
      ptr = filebuf+i+1;
    }
  }
  ptr     = filebuf;
  lastptr = filebuf + statbuf.st_size;
}

void closeFile ()
{
  free (lines);
  free (filebuf);
}

static int line = 1;
static int pos = -1;
static int lastchar = -1;
static int lastfoo = -1;

static int readch ()
{
  int ch;

  lastchar = lastfoo;
  ch = *ptr++;
  if ( ptr == lastptr ) return (lastfoo=EOF);

  if (ch == '\0') {
    ch = '\n';
    line++;
    pos = -1;
  }
  else
    pos++;

  return (lastfoo=ch);
}

static void gotopos (int l, int p)
{
  int i;

  line = l;
  pos  = -1;
  ptr  = lines[l-1];
  for (i=0; i<p; i++) readch();
}

char *getLine (int l)
{
  return lines[l-1];
}

static char resbuf[80];

static int body_end_line = -1;
static int look_next_line = -1;
static int look_next_pos  = -1;

void findFuncOrClass (int start, char **result)
{
  int ch;
  int braces = 1;
  int last = 0;
  int header_start_line = -1;
  int header_start_pos  = -1;
  int header_end_line = -1;
  int header_end_pos  = -1;
  int hline = -1;
  int hpos  = -1;
  int body_start_line = -1;
  int body_start_pos  = -1;

  gotopos (start, 0);

  while ( (ch = readch()) != EOF ) {
    if ( header_start_line == -1 ) {

      /* -- look for start of function or class header ------------------- */
      if ( ((ch >= 'a') && (ch <= 'z')) || ((ch >= 'A') && (ch <= 'Z')) ) {
        header_start_line = line;
        header_start_pos  = pos;
      }
    }
    else if ( header_end_line == -1 ) {

      /* -- look for end of function or class header --------------------- */
      if ( (ch != '{') && (ch != ' ') && (ch != '\t') &&
           (ch != '\n') && (ch != '\f') && (ch != '\r') ) {
        hline = line;
        hpos  = pos;
      }
      else if ( ch == '{' ) {
        header_end_line = hline;
        header_end_pos  = hpos;
        body_start_line = line;
        body_start_pos  = pos;

        /* -- for findNextChild -- */
        look_next_line  = line;
        look_next_pos   = pos;
      }
    }
    else {

      /* -- look for end of function or class body ----------------------- */
      if ( ch == '/' ) {
        if ( (ch = readch()) == '/' ) {
          while ( (ch = readch()) != '\n' ) /*skip*/ ;
        }
        else if ( ch == '*' ) {
          last = 0;
          do {
            if ( ((ch = readch()) == '/') && (last == '*') )
              break;
            last = ch;
          }
          while ( ch != EOF );
        }
      }

      if ( ch == '\'' ) {
        do {
          if ( ((ch = readch()) == '\\') && ((ch = readch()) == '\'') )
            ch = readch();
        }
        while ( ch != '\'' );
      }
      else if ( ch == '\"' ) {
        do {
          if ( (ch = readch()) == '\\' ) {
            if ( (ch = readch()) == '\"' )
              ch = readch();
          }
        }
        while ( ch != '\"' );
      }
      else if ( ch == '{' ) {
        braces++;
      }
      else if ( ch == '}' ) {
        braces--;
        if ( braces == 0 ) {
          sprintf (resbuf, "@ %d.%d %d.%d %d.%d %d.%d",
                  header_start_line, header_start_pos,
                  header_end_line,   header_end_pos + 1,
                  body_start_line,   body_start_pos,
                  line, pos + 1);
          *result = resbuf;

          /* -- for findNextChild -- */
          body_end_line = line;
          return;
        }
      }
    }
  }
  sprintf (resbuf, "@ - - - -");
  *result = resbuf;
}

static void findNext (char *child, char **result, int marknewpos,
                      int start_line, int start_pos)
{
  int ch;
  int last = 0;
  int child_start_line = -1;
  int child_start_pos  = -1;
  char *pattern;

  if ( ((*child >= 'a') && (*child <= 'z')) ||
       ((*child >= 'A') && (*child <= 'Z')) ) {
    gotopos (start_line, start_pos);
    pattern = child;
    while ( ((ch = readch()) != EOF) && (line <= body_end_line) ) {
      if ( ch == '/' ) {
        if ( (ch = readch()) == '/' ) {
          while ( (ch = readch()) != '\n' ) /*skip*/ ;
        }
        else if ( ch == '*' ) {
          last = 0;
          do {
            if ( ((ch = readch()) == '/') && (last == '*') )
            break;
            last = ch;
          }
          while ( ch != EOF );
        }
      }

      if ( ch == '\'' ) {
        do {
          if (((ch = readch()) == '\\') && ((ch = readch()) == '\''))
            ch = readch();
        }
        while ( ch != '\'' );
      }
      else if ( ch == '\"' ) {
        do {
          if ( (ch = readch()) == '\\' ) {
            if ( (ch = readch()) == '\"' )
              ch = readch();
          }
        }
        while ( ch != '\"' );
      }
      else if ( ch == *pattern ) {
        if ( child_start_line == -1) {
          if ( ((lastchar >= 'a') && (lastchar <= 'z')) ||
               ((lastchar >= 'A') && (lastchar <= 'Z')) ) {
            continue;
          }
          child_start_line = line;
          child_start_pos  = pos;
        }
        pattern++;
        if ( *pattern == '\0' ) {
          sprintf (resbuf, "%d.%d %d.%d",
                   child_start_line, child_start_pos,
                   line, pos + 1);
          *result = resbuf;
          if ( marknewpos ) {
            look_next_line = line;
            look_next_pos  = pos;
          }
          return;
        }
      }
      else {
        pattern = child;
        child_start_line = -1;
      }
    }
  }

  /* -- not found or non-alphabetic name -- */
  sprintf (resbuf, "- -");
  *result = resbuf;
}

void findNextChild (char *child, char **result, int marknewpos)
{
  findNext (child, result, marknewpos, look_next_line, look_next_pos);
}

void findNextMember (char *name, int line, char **result)
{
  static char nbuf[12];

  if ( ((*name >= 'a') && (*name <= 'z')) ||
       ((*name >= 'A') && (*name <= 'Z')) )
    findNext (name, result, FALSE, line, 0);
  else {
    sprintf (nbuf, "operator%s", name);
    findNext (nbuf, result, FALSE, line, 0);
  }
}

# ifdef EBUG

int main (int argc, char **argv)
{
  int i;
  char *r;

  openFile (argv[1]);
  findFuncOrClass (atoi(argv[2]), &r);
  printf ("%s: %s\n", argv[1], r);
  for (i=3; i<argc; i++) {
    findNextChild (argv[i], &r, TRUE);
    printf ("%s: %s\n", argv[i], r);
  }
  closeFile ();
  exit (0);
}
# endif
